from functools import reduce
import logging
import operator
import collections
import six

import numpy as np
import pyarrow as pa
from vaex.dataframe import DataFrame
from vaex.docstrings import docsubst

import vaex.array_types
from vaex.delayed import delayed_args, delayed_dict, delayed_list
from vaex.utils import _ensure_list, _ensure_string_from_expression
import vaex
import vaex.hash
import vaex.utils

try:
    collections_abc = collections.abc
except AttributeError:
    collections_abc = collections

__all__ = ['GroupBy', 'Grouper', 'BinnerTime']

logger = logging.getLogger("vaex.groupby")
_USE_DELAY = True
_EXPERIMENTAL_BINNER_HASH = False


# pure Python to avoid int overflow
product = lambda l: reduce(operator.mul, l, 1)


class BinnerBase:
    simpler = None  # a binner may realize there is a simpler fallback
    def extract_center(self, dim, ar):
        # gets rid of the nan and out of bound values
        slices = [
            slice(None, None),
        ] * vaex.array_types.ndim(ar)
        slices[dim] = slice(2, -1)
        return vaex.array_types.getitem(ar, tuple(slices))

class Binner(BinnerBase):
    dense = False
    def __init__(self, expression, vmin, vmax, bins, df=None, label=None):
        self.df = df or expression.df
        self.expression = self.df[str(expression)]
        self.label = label or self.expression._label
        self.vmin = vmin
        self.vmax = vmax
        self.N = bins
        self.binby_expression = str(expression)
        self.bin_values = self.df.bin_centers(expression, (self.vmin, self.vmax), bins)
        self.sort_indices = None
        self._promise = vaex.promise.Promise.fulfilled(None)

    def _create_binner(self, df):
        assert (df.dataset == self.df.dataset), "you passed a dataframe with a different dataset to the grouper/binned"
        self.df = df
        self.binner = self.df._binner_scalar(self.binby_expression, (self.vmin, self.vmax), self.N)

class BinnerTime(BinnerBase):
    """Bins an expression in a specified datetime resolution.

    Note that different with :class:`Grouper`, all values between the minimum and maximum
    datetime (incremented by the resolution step) will be present in the dataframe.
    A normal groupby would only contain datatime values that are present in the original
    dataframe.

    Example:

    >>> import vaex
    >>> import numpy as np
    >>> t = np.arange('2015-01-01', '2015-02-01', dtype=np.datetime64)
    >>> y = np.arange(len(t))
    >>> df = vaex.from_arrays(t=t, y=y)
    >>> df.groupby(vaex.BinnerTime.per_week(df.t)).agg({'y' : 'sum'})
      #  t             y
      0  2015-01-01   21
      1  2015-01-08   70
      2  2015-01-15  119
      3  2015-01-22  168
      4  2015-01-29   87

    """

    dense = False

    def __init__(self, expression, resolution='W', df=None, every=1):
        self.resolution = resolution
        self.expression = expression
        self.df = df or expression.ds
        self.every = every
        self.sort_indices = None
        # make sure it's an expression
        self.expression = self.df[str(self.expression)]
        self.label = self.expression._label
        self.tmin, self.tmax = self.df[str(self.expression)].minmax()

        self.resolution_type = 'M8[%s]' % self.resolution
        dt = (self.tmax.astype(self.resolution_type) - self.tmin.astype(self.resolution_type))
        self.N = (dt.astype(int).item() + 1)
        # divide by every, and round up
        self.N = (self.N + every - 1) // every
        self.bin_values = np.arange(self.tmin.astype(self.resolution_type), self.tmax.astype(self.resolution_type)+1, every)
        self._promise = vaex.promise.Promise.fulfilled(None)

    def _create_binner(self, df):
        # TODO: we modify the dataframe in place, this is not nice
        assert df.dataset == self.df.dataset, "you passed a dataframe with a different dataset to the grouper/binned"
        self.df = df
        self.begin_name = df.add_variable('t_begin', self.tmin.astype(self.resolution_type), unique=True)
        # TODO: import integer from future?
        self.binby_expression = str(df['%s - %s' % (self.expression.astype(self.resolution_type), self.begin_name)].astype('int') // self.every)
        self.binner = df._binner_ordinal(self.binby_expression, self.N)

    @classmethod
    def per_day(cls, expression, df=None):
        return cls(expression, 'D', df)

    @classmethod
    def per_week(cls, expression, df=None):
        return cls(expression, 'W', df)

    @classmethod
    def per_month(cls, expression, df=None):
        return cls(expression, 'M', df)

    @classmethod
    def per_quarter(cls, expression, df=None, every=1):
        return cls(expression, 'M', df, every=3*every)

    @classmethod
    def per_year(cls, expression, df=None):
        return cls(expression, 'Y', df)

    def extract_center(self, dim, ar):
        # gets rid of the nan and out of bound values
        slices = [
            slice(None, None),
        ] * vaex.array_types.ndim(ar)
        slices[dim] = slice(0, -2) # remove null and nan
        return vaex.array_types.getitem(ar, tuple(slices))

class BinnerInteger(BinnerBase):
    '''Bins an expression into it's natural bin (i.e. 5 for the number 5)

    Useful for boolean, int8/uint8, which only have a limited number of possibilities (2, 256)
    '''
    # these are always true
    pre_sort = True
    sort = True

    def __init__(self, expression, label=None, dropmissing=False, min_value=None, max_value=None, dense=False, sort=False, ascending=True):
        self.expression = expression
        self.df = expression.df
        self.dtype = self.expression.dtype
        self.label = label or self.expression._label
        self.min_value = 0
        self.dropmissing = dropmissing
        def make_array_with_null(vmin, vmax, dtype='int64'):
            if sort and not ascending:
                values = np.arange(vmax - 1, vmin - 2, -1, dtype=dtype)
            else:
                values = np.arange(vmin, vmax + 1, dtype=dtype)
            mask = np.zeros(values.shape, dtype="?")
            mask[len(values) - 1] = 1
            values[len(values) - 1] = 0
            return np.ma.array(values, mask=mask, shrink=False)

        self.dense = dense
        if self.dtype.numpy == np.dtype('bool'):
            if sort and not ascending:
                self.bin_values = vaex.array_types.to_numpy(pa.array([False, True, None]))
            else:
                self.bin_values = vaex.array_types.to_numpy(pa.array([False, True, None]))
            self.N = 2
        elif self.dtype.numpy == np.dtype('uint8'):
            self.bin_values = make_array_with_null(0, 256)
            self.N = 256
        elif self.dtype.numpy == np.dtype('int8'):
            self.min_value = -128
            self.bin_values = make_array_with_null(-128, 128)
            self.N = 256
        elif min_value is not None and max_value is not None:
            self.min_value = min_value
            self.N = int(max_value - min_value + 1)
            self.bin_values = make_array_with_null(min_value, max_value + 1)
        else:
            raise TypeError(f"Only boolean, int8 and uint8 are supported, not {self.dtype}, or private min_value and max_value")
        if self.dropmissing:
            # if we remove the missing values, we are already sorted
            self.bin_values = self.bin_values[:-1]
        self.binby_expression = str(self.expression)
        # no need to invert if we don't sort, but sort is always implicit
        self.invert = sort and not ascending
        if self.invert:
            # invert the range, taking into account min_value
            self.combine_expression = (self.N - 1 - (self.df[self.binby_expression] - self.min_value) + self.min_value).fillmissing(self.N).expression
        else:
            self.combine_expression = self.df[self.binby_expression].fillmissing(self.N).expression
        # self.combine_expression.expression = self.combine_expression.expression
        self.sort_indices = None
        self._promise = vaex.promise.Promise.fulfilled(None)

    def extract_center(self, dim, ar):
        # gets rid of the nan and out of bound values
        # indices = n
        slices = [
            slice(None, None),
        ] * vaex.array_types.ndim(ar)
        if self.dropmissing:
            slices[dim] = slice(0, -2)  # remove null and nan
        else:
            slices[dim] = slice(0, -1)  # remove nan
        return vaex.array_types.getitem(ar, tuple(slices))

    def _create_binner(self, df: DataFrame):
        assert df.dataset == self.df.dataset, "you passed a dataframe with a different dataset to the grouper/binned"
        self.df = df
        self.binner = self.df._binner_ordinal(self.binby_expression, self.N, self.min_value, self.invert)


class Grouper(BinnerBase):
    """Bins an expression to a set of unique bins, like an SQL like groupby."""

    dense = True

    def __init__(self, expression, df=None, sort=False, ascending=True, pre_sort=True, row_limit=None, df_original=None, materialize_experimental=False, progress=None, allow_simplify=False):
        self.df = df or expression.ds
        self.sort = sort
        self.pre_sort = pre_sort
        # we prefer to calculate the set the original dataframe to have better cache hits, and modify df
        if df_original is None:
            df_original = self.df
        self.sort = sort
        self.expression = expression
        self.allow_simplify = allow_simplify
        # make sure it's an expression
        self.expression = self.df[_ensure_string_from_expression(self.expression)]
        self.label = self.expression._label
        self.progressbar = vaex.utils.progressbars(progress, title=f"grouper: {repr(self.label)}" )
        dtype = self.expression.dtype
        if materialize_experimental:
            set, values = df_original._set(self.expression, limit=row_limit, return_inverse=True)
            # TODO: add column should have a unique argument
            self.df.add_column(f'__materialized_{self.label}', values)

            self.bin_values = set.key_array()
            if isinstance(self.bin_values, vaex.superstrings.StringList64):
                self.bin_values = pa.array(self.bin_values.to_numpy())
            self.binby_expression = 'bla'
            self.N = len(self.bin_values)
            self.min_value = 0
            self.binner = self.df._binner_ordinal('bla', self.N, self.min_value)
            self.sort_indices = None
        else:
            @vaex.delayed
            def process(hashmap_unique: vaex.hash.HashMapUnique):
                self.bin_values = hashmap_unique.keys()
                if self.allow_simplify and dtype == int and len(self.bin_values):
                    vmin = self.bin_values.min()
                    vmax = self.bin_values.max()
                    int_range = vmax - vmin + 1
                    # we allow for 25% unused 'slots'
                    bins = len(self.bin_values)
                    if int_range <= (bins * 4 / 3):
                        dense = bins == int_range
                        self.simpler = BinnerInteger(self.expression, min_value=vmin, max_value=vmax, dropmissing=not hashmap_unique.has_null, dense=dense, sort=sort, ascending=ascending)
                        return

                if vaex.dtype_of(self.bin_values) == int and len(self.bin_values):
                    min_value, max_value = self.bin_values.min(), self.bin_values.max()
                    self.bin_values = self.bin_values.astype(vaex.utils.required_dtype_for_range(min_value, max_value))
                logger.debug('Constructed grouper for expression %s with %i values', str(expression), len(self.bin_values))

                if self.sort:
                    if pre_sort:
                        hashmap_unique, self.bin_values = hashmap_unique.sorted(keys=self.bin_values, ascending=ascending, return_keys=True)
                        self.sort_indices = None
                    else:
                        indices = pa.compute.sort_indices(self.bin_values, sort_keys=[("x", "ascending" if ascending else "descending")])
                        self.sort_indices = vaex.array_types.to_numpy(indices)
                        # the bin_values will still be pre sorted, maybe that is confusing (implementation detail)
                        self.bin_values = pa.compute.take(self.bin_values, self.sort_indices)
                else:
                    self.sort_indices = None
                self.hashmap_unique = hashmap_unique

                self.basename = 'hashmap_unique_%s' % vaex.utils._python_save_name(str(self.expression) + "_" + hashmap_unique.fingerprint)

                self.N = len(self.bin_values)
                # for datetimes, we converted to int
                if dtype.is_datetime:
                    self.bin_values = dtype.create_array(self.bin_values)
            self._promise = process(df_original._hash_map_unique(self.expression, limit=row_limit, delay=True, progress=self.progressbar))

    def __repr__(self):
        return f"vaex.groupby.Grouper({str(self.expression)})"

    def _create_binner(self, df):
        # TODO: we modify the dataframe in place, this is not nice
        assert df.dataset == self.df.dataset, "you passed a dataframe with a different dataset to the grouper/binned"
        self.df = df
        if self.basename not in self.df.variables:
            self.hash_map_unique_name = df.add_variable(self.basename, self.hashmap_unique, unique=True)
        else:
            self.hash_map_unique_name = self.basename
        if _EXPERIMENTAL_BINNER_HASH:
            self.binby_expression = str(self.expression)
            self.binner = self.df._binner_hash(self.binby_expression, self.hashmap_unique)
        else:
            self.binby_expression = "_ordinal_values(%s, %s)" % (self.expression, self.hash_map_unique_name)
            self.binner = self.df._binner_ordinal(self.binby_expression, self.N)
        self.combine_expression = self.binby_expression

    def extract_center(self, dim, ar):
        slices = [
            slice(None, None),
        ] * vaex.array_types.ndim(ar)
        if _EXPERIMENTAL_BINNER_HASH:
            slices[dim] = slice(1, -1)
        else:
            slices[dim] = slice(0, -2)  # remove null and nan, actually null in grouper is part of the 'dictionary' (self.bin_values)
        return vaex.array_types.getitem(ar, tuple(slices))


class GrouperCombined(Grouper):
    dense = True

    def __init__(self, expression, df, multipliers, parents, sort, row_limit=None, progress=None):
        '''Will group by 1 expression, which is build up from multiple expressions.

        Used in the sparse/combined group by.
        '''
        super().__init__(expression, df, sort=sort, row_limit=row_limit, progress=progress, allow_simplify=False)
        assert len(multipliers) == len(parents)
        self.parents = parents

        assert multipliers[-1] == 1
        self.df = df
        for parent in parents:
            assert isinstance(parent, (Grouper, GrouperCategory, BinnerInteger, GrouperLimited)), "only (Grouper, GrouperCategory) supported for combining"
        self.label = 'SHOULD_NOT_BE_USED'
        self.expression = expression
        # efficient way to find the original bin values (parent.bin_value) from the 'compressed'
        # self.bin_values
        progressbar = self.progressbar.add("extract labels from sparse set")
        @vaex.delayed
        def process(_ignore):
            logger.info(f"extracing indices of parent groupers ({self.N:,} unique rows)")
            df = vaex.from_dict({'row': vaex.vrange(0, self.N, dtype='i8'), 'bin_value': self.bin_values})
            df[f'index_0'] = df['bin_value'] // multipliers[0]
            df[f'leftover_0'] = df[f'bin_value'] % multipliers[0]
            for i in range(1, len(multipliers)):
                df[f'index_{i}'] = df[f'leftover_{i-1}'] // multipliers[i]
                df[f'leftover_{i}'] = df[f'leftover_{i-1}'] % multipliers[i]
            columns = [f'index_{i}' for i in range(len(multipliers))]
            indices_parents = df.evaluate(columns, progress=progressbar)
            def compress(ar):
                if vaex.dtype_of(ar).kind == 'i':
                    ar = vaex.array_types.to_numpy(ar)
                    max_value = ar.max()
                    ar = ar.astype(vaex.utils.required_dtype_for_max(max_value))
                    return ar
            indices_parents = [compress(ar) for ar in indices_parents]
            bin_values = {}
            logger.info(f"extracing labels of parent groupers...")
            # NOTE: we can also use dict encoding instead of take
            for indices, parent in zip(indices_parents, parents):
                if sort:
                    assert parent.pre_sort, "cannot sort while parent not presorted"
                    assert parent.sort_indices is None
                dtype = vaex.dtype_of(parent.bin_values)
                if dtype.is_struct:
                    # collapse parent struct into our flat struct
                    for field, ar in zip(parent.bin_values.type, parent.bin_values.flatten()):
                        bin_values[field.name] = ar.take(indices)
                        # bin_values[field.name] = pa.DictionaryArray.from_arrays(indices, ar)
                else:
                    bin_values[parent.label] = parent.bin_values.take(indices)
                    # bin_values[parent.label] = pa.DictionaryArray.from_arrays(indices, parent.bin_values)
            logger.info(f"extracing labels of parent groupers done")
            return pa.StructArray.from_arrays(bin_values.values(), bin_values.keys())
        self._promise_parent = self._promise
        def key_function():
            fp = vaex.cache.fingerprint(expression.fingerprint())
            return f'groupby-combined-{fp}'
        lookup_originals = vaex.cache._memoize(process, key_function=key_function, delay=True)
        @vaex.delayed
        def set(bin_values):
            self.bin_values = bin_values
        self._promise = set(lookup_originals(self._promise_parent))


class GrouperCategory(BinnerBase):
    """Faster grouper that will use the fact that a column is categorical."""

    dense = True

    def __init__(self, expression, df=None, sort=False, ascending=True, row_limit=None, pre_sort=True):
        self.df = df or expression.ds
        self.sort = sort
        self.pre_sort = pre_sort
        # make sure it's an expression
        expression = self.df[str(expression)]
        self.expression_original = expression
        self.label = expression._label
        self.expression = expression.index_values() if expression.dtype.is_encoded else expression
        self.row_limit = row_limit

        self.min_value = self.df.category_offset(self.expression_original)
        self.bin_values = self.df.category_labels(self.expression_original, aslist=False)
        self.N = self.df.category_count(self.expression_original)
        dtype = self.expression.dtype  # should be the dtype of the 'codes'
        if self.sort:
            # not pre-sorting is faster
            sort_indices = pa.compute.sort_indices(self.bin_values, sort_keys=[("x", "ascending" if ascending else "descending")])
            self.bin_values = pa.compute.take(self.bin_values, sort_indices)
            if self.pre_sort:
                # we will map from int to int
                sort_indices = vaex.array_types.to_numpy(sort_indices)
                fingerprint = self.expression.fingerprint() + "-grouper-sort-mapper"
                self.hash_map_unique = vaex.hash.HashMapUnique.from_keys(sort_indices + self.min_value, fingerprint=fingerprint)
                self.min_value = 0
                self.sort_indices = None
                self.basename = "hash_map_unique_%s" % vaex.utils._python_save_name(str(self.expression) + "_" + self.hash_map_unique.fingerprint)
            else:
                self.sort_indices = sort_indices
        else:
            self.sort_indices = None
        if isinstance(self.bin_values, list):
            self.bin_values = pa.array(self.bin_values)

        if row_limit is not None:
            if self.N > row_limit:
                raise vaex.RowLimitException(f'Resulting grouper has {self.N:,} unique combinations, which is larger than the allowed row limit of {row_limit:,}')
        # TODO: what do we do with null values for categories?
        # if self.set.has_null:
        #     self.N += 1
        #     keys += ['null']
        self._promise = vaex.promise.Promise.fulfilled(None)

    def _create_binner(self, df):
        assert df.dataset == self.df.dataset, "you passed a dataframe with a different dataset to the grouper/binned"
        self.df = df
        if self.sort and self.pre_sort:
            if self.basename not in self.df.variables:
                self.var_name = df.add_variable(self.basename, self.hash_map_unique, unique=True)
            else:
                self.var_name = self.basename
            self.binby_expression = "_ordinal_values(%s, %s)" % (self.expression, self.var_name)
            self.binner = self.df._binner_ordinal(self.binby_expression, self.N, 0)
        else:
            self.binby_expression = str(self.expression)
            self.binner = self.df._binner_ordinal(self.binby_expression, self.N, self.min_value)
        self.combine_expression = self.binby_expression

    def extract_center(self, dim, ar):
        slices = [
            slice(None, None),
        ] * vaex.array_types.ndim(ar)
        slices[dim] = slice(0, -2)  # again, null is in the dictionary
        return vaex.array_types.getitem(ar, tuple(slices))

class GrouperLimited(BinnerBase):
    """Group to a limited set of values, store the rest in an (optional) other bin"""

    dense = True

    def __init__(self, expression, values, keep_other=True, other_value=None, sort=False, ascending=True, label=None, df=None):
        self.df = df or expression.df
        self.sort = sort
        self.pre_sort = True
        self.expression = self.df[str(expression)]
        self.label = label or self.expression._label
        self.keep_other = keep_other
        if isinstance(values, pa.ChunkedArray):
            values = pa.concat_arrays(values.chunks)
        if sort:
            indices = pa.compute.sort_indices(values, sort_keys=[("x", "ascending" if ascending else "descending")])
            values = pa.compute.take(values, indices)

        if self.keep_other:
            self.bin_values = pa.array(vaex.array_types.tolist(values) + [other_value])
            self.values = self.bin_values.slice(0, len(self.bin_values) - 1)
        else:
            raise NotImplementedError("not supported yet")
            # although we can support this, it will fail with _combine, because of
            # the mapping of the set to -1
            self.bin_values = pa.array(vaex.array_types.tolist(values))
            self.values = self.bin_values
        self.N = len(self.bin_values)
        fp = vaex.cache.fingerprint(values)
        fingerprint = f"set-grouper-fixed-{fp}"
        self.hash_map_unique = vaex.hash.HashMapUnique.from_keys(self.values, fingerprint=fingerprint)

        self.basename = "hash_map_unique_%s" % vaex.utils._python_save_name(str(self.expression) + "_" + self.hash_map_unique.fingerprint)
        self.sort_indices = None
        self._promise = vaex.promise.Promise.fulfilled(None)

    def _create_binner(self, df):
        assert df.dataset == self.df.dataset, "you passed a dataframe with a different dataset to the grouper/binned"
        self.df = df
        if self.basename not in self.df.variables:
            self.var_name = df.add_variable(self.basename, self.hash_map_unique, unique=True)
        else:
            self.var_name = self.basename
        # modulo N will map -1 (value not found) to N-1
        self.binby_expression = "_ordinal_values(%s, %s) %% %s" % (self.expression, self.var_name, self.N)
        self.combine_expression = self.binby_expression
        self.binner = self.df._binner_ordinal(self.binby_expression, self.N)

    def extract_center(self, dim, ar):
        # gets rid of the nan and out of bound values
        # indices = n
        slices = [
            slice(None, None),
        ] * vaex.array_types.ndim(ar)
        slices[dim] = slice(0, -2)  # remove null values
        return vaex.array_types.getitem(ar, tuple(slices))


def _combine(df, groupers, sort, row_limit=None, progress=None):
    for grouper in groupers:
        if isinstance(grouper, Binner):
            raise NotImplementedError('Cannot combined Binner with other groupers yet')

    progressbar = vaex.utils.progressbars(progress, title="find sparse entries / compress")
    groupers = groupers.copy()
    max_count_64bit = 2**63-1
    first = groupers.pop(0)
    combine_now = [first]
    combine_later = []
    counts = [first.N]

    for grouper in groupers:
        assert isinstance(grouper, (Grouper, GrouperCategory, BinnerInteger, GrouperLimited)), f"not supported binner {type(grouper)}"

    # when does the cartesian product overflow 64 bits?
    next = groupers.pop(0)
    while (product(counts) * next.N < max_count_64bit):
        counts.append(next.N)
        combine_now.append(next)
        if groupers:
            next = groupers.pop(0)
        else:
            next = None
            break
    counts.append(1)
    # decreasing [40, 4, 1] for 2 groupers (N=10 and N=4)
    cumulative_counts = np.cumproduct(counts[::-1], dtype='i8').tolist()[::-1]
    assert len(combine_now) >= 2
    combine_later = ([next] if next else []) + groupers

    logger.info("combining groupers, now %r, next %r", combine_now, combine_later)
    combine_expressions = [df[k.combine_expression] for k in combine_now]
    for i in range(0, len(combine_expressions)):
        combine_expression = combine_expressions[i]
        dtype = vaex.utils.required_dtype_for_max(cumulative_counts[i])
        if isinstance(combine_now[i], (GrouperCategory, BinnerInteger)) and combine_now[i].min_value != 0:
            combine_expression -= combine_now[i].min_value
        combine_expression = combine_expression.astype(str(dtype))
        if cumulative_counts[i + 1] != 1:
            combine_expression = combine_expression * cumulative_counts[i + 1]
        combine_expressions[i] = combine_expression
    expression = reduce(operator.add, combine_expressions)
    grouper = GrouperCombined(expression, df, multipliers=cumulative_counts[1:], parents=combine_now, sort=sort, row_limit=row_limit, progress=progressbar)
    if combine_later:
        @vaex.delayed
        def combine(_ignore):
            # recursively add more of the groupers (because of 64 bit overflow)
            # return 1
            grouper._create_binner(df)
            new_grouper = _combine(df, [grouper] + combine_later, sort=sort)
            return new_grouper

        return combine(grouper._promise)
    return grouper._promise.then(lambda x: grouper)


def grouper(df, column, sort=False, ascending=True, row_limit=None, df_original=None, progress=None):
    if df_original is None:
        df_original = df
    expression = df[_ensure_string_from_expression(column)]
    if df.is_category(column):
        by_value = GrouperCategory(expression, sort=sort, ascending=ascending, row_limit=row_limit)
    else:
        dtype = expression.dtype
        if dtype == np.dtype("uint8") or dtype == np.dtype("int8") or dtype == np.dtype("bool"):
            by_value = BinnerInteger(expression, sort=sort, ascending=ascending)  # always sorted, and pre_sorted
        else:
            # we cannot mix _combine with BinnerInteger yet
            by_value = Grouper(expression, sort=sort, ascending=ascending, row_limit=row_limit, df_original=df_original, progress=progress, allow_simplify=True)
    return by_value

class GroupByBase(object):
    def __init__(self, df, by, sort=False, ascending=True, combine=False, expand=True, row_limit=None, copy=True, progress=None):
        '''Note that row_limit only works in combination with combine=True'''
        df_original = df
        if copy:
            df = df.copy() # we will mutate the df (Add variables), this will keep the original dataframe unchanged
        self.df = df
        self.expand = expand  # keep as pyarrow struct?
        self.progressbar = vaex.utils.progressbars(progress)
        self.progressbar_groupers = self.progressbar.add("groupers")

        if by is None:
            by = []
        elif not isinstance(by, collections_abc.Iterable)\
            or isinstance(by, six.string_types):
            by = [by]

        if isinstance(ascending, (list, tuple)):
            assert len(ascending) == len(by), 'If "ascending" is a list, it must have the same number of elements as "by".'
        else:
            ascending = [ascending] * len(by)

        if isinstance(sort, (list, tuple)):
            assert len(sort) == len(by), 'If "sort" is a list, it must have the same number of elements as "by".'
        else:
            sort = [sort] * len(by)
        self.sort = sort
        self.ascending = ascending

        self.by = []
        self.by_original = by
        for by_value, sort, ascending in zip(by, self.sort, self.ascending):
            if not isinstance(by_value, BinnerBase):
                expression = df[_ensure_string_from_expression(by_value)]
                by_value = grouper(df, by_value, row_limit=row_limit, df_original=df_original, progress=self.progressbar_groupers, sort=sort, ascending=ascending)
            self.by.append(by_value)
        @vaex.delayed
        def possible_combine(*binner_promises):
            # if a binner realized there is a simpler way (e.g. grouper -> intbinner)
            self.by = [by.simpler if by.simpler is not None else by for by in self.by]
            # because binners can be created from other dataframes (we make a copy)
            # we let it mutate *our* dataframe
            for binner in self.by:
                binner._create_binner(self.df)
            cells = product([grouper.N for grouper in self.by])
            @vaex.delayed
            def set_combined(combined):
                combined._create_binner(self.df)
                self.by = [combined]
                self.combine = True
            if ((row_limit is not None) or (combine is True)) and len(self.by) >= 2 and cells > 0:
                promise = set_combined(_combine(self.df, self.by, sort=sort, row_limit=row_limit, progress=self.progressbar_groupers))
            elif combine == 'auto' and len(self.by) >= 2:
                # default assume we cannot combined
                self.combine = False
                promise = vaex.promise.Promise.fulfilled(None)
                # don't even try when one grouper has 0 options
                if cells > 0:
                    dim = len(self.by)
                    rows = df.length_unfiltered()  # we don't want to trigger a computation
                    occupancy = rows/cells
                    logger.debug('%s rows and %s grid cells => occupancy=%s', rows, cells, occupancy)
                    # we want each cell to have a least 10x occupacy
                    if occupancy < 10:
                        logger.info(f'Combining {len(self.by)} groupers into 1')
                        promise = set_combined(_combine(self.df, self.by, sort=sort, row_limit=row_limit, progress=self.progressbar_groupers))
                        self.combine = True
            else:
                self.combine = False
                promise = vaex.promise.Promise.fulfilled(None)
            @vaex.delayed
            def process(_ignore):
                self.dense = len(self.by) == 1 and self.by[0].dense
                self.groupby_expression = [str(by.expression) for by in self.by]
                self.binners = tuple(by.binner for by in self.by)
                self.shape = [by.N for by in self.by]
                self.dims = [by.label for by in self.by]
                self.progressbar_groupers(1)
            return process(promise)

        self._promise_by = self.progressbar_groupers.exit_on(possible_combine(*[by._promise for by in self.by]))

    @property
    def _coords1d(self):
        return [k.bin_values for k in self.by]

    def _agg(self, actions, progressbar_agg):
        df = self.df
        if isinstance(actions, collections_abc.Mapping):
            actions = list(actions.items())
        elif not isinstance(actions, collections_abc.Iterable)\
            or isinstance(actions, six.string_types):
            actions = [actions]


        grids = {}
        self.counts = None  # if we already do count *, we'd like to know so we don't have to do it ourselves!

        def add(aggregate, column_name=None, override_name=None):
            if column_name is None or override_name is not None:
                column_name = aggregate.pretty_name(override_name, df)
            aggregate.edges = True  # is this ok to override?
            values = df._agg(aggregate, self.binners, delay=_USE_DELAY, progress=progressbar_agg)
            grids[column_name] = values
            if isinstance(aggregate, vaex.agg.AggregatorDescriptorBasic)\
                and aggregate.name == 'AggCount'\
                and aggregate.expressions == []\
                and (aggregate.selection is None or aggregate.selection is False):
                self.counts = values

        for item in actions:
            override_name = None
            if isinstance(item, tuple):
                name, aggregates = item
            else:
                aggregates = item
                name = None

            if not isinstance(aggregates, collections_abc.Iterable)\
                or isinstance(aggregates, six.string_types):
                # not a list, or a string
                aggregates = [aggregates]
            else:
                # if we have a form of {'z': [.., ...]}
                # we want to have a name like z_<agg name>
                if name is not None:
                    override_name = name
            for aggregate in aggregates:
                if isinstance(aggregate, six.string_types) and aggregate == "count":
                    add(vaex.agg.count(), 'count' if name is None else name)
                else:
                    if isinstance(aggregate, six.string_types):
                        aggregate = vaex.agg.aggregates[aggregate]
                    if callable(aggregate):
                        if name is None:
                            # we use all columns
                            for column_name in df.get_column_names():
                                if column_name not in self.groupby_expression:
                                    add(aggregate(column_name), override_name=override_name)
                        else:
                            add(aggregate(name), name, override_name=override_name)
                    else:
                        add(aggregate, name, override_name=override_name)
        return grids

    @property
    def groups(self):
        for group, df in self:
            yield group

    def get_group(self, group):
        '''Get DataFrame containing a single group from the :class:`GroupBy`.

        Example:

        >>> import vaex
        >>> import vaex.datasets
        >>> df = vaex.datasets.titanic()
        >>> g1 = df.groupby(by='pclass')
        >>> df_group1 = g1.get_group(1)
        >>> df_group1.head(3)
          #    pclass  survived    name                            sex         age    sibsp    parch    ticket     fare  cabin    embarked    boat      body  home_dest
          0         1  True        Allen, Miss. Elisabeth Walton   female  29             0        0     24160  211.338  B5       S           2          nan  St Louis, MO
          1         1  True        Allison, Master. Hudson Trevor  male     0.9167        1        2    113781  151.55   C22 C26  S           11         nan  Montreal, PQ / Chesterville, ON
          2         1  False       Allison, Miss. Helen Loraine    female   2             1        2    113781  151.55   C22 C26  S           --         nan  Montreal, PQ / Chesterville, ON

        >>> df = vaex.datasets.titanic()
        >>> g2 = df.groupby(by=['pclass', 'sex'])
        >>> df_group2 = g2.get_group([1, 'female'])
        >>> df_group2.head(3)
          #    pclass  survived    name                                             sex       age    sibsp    parch    ticket     fare  cabin    embarked    boat      body  home_dest
          0         1  True        Allen, Miss. Elisabeth Walton                    female     29        0        0     24160  211.338  B5       S           2          nan  St Louis, MO
          1         1  False       Allison, Miss. Helen Loraine                     female      2        1        2    113781  151.55   C22 C26  S           --         nan  Montreal, PQ / Chesterville, ON
          2         1  False       Allison, Mrs. Hudson J C (Bessie Waldo Daniels)  female     25        1        2    113781  151.55   C22 C26  S           --         nan  Montreal, PQ / Chesterville, ON

        :param group: A value or a list of values of the expressions used to create the groupby object.
            If `assume_sparse=True` when greating the groupby object, this param takes an int corresponding to the particular groupby combination.
        :rtype: DataFrame
        '''
        if self.combine:
            assert isinstance(group, int)
            filter_expression = self.df[str(self.by[0].binby_expression)] == group
        else:
            group = _ensure_list(group)
            values = group
            filter_expressions = [self.df[expression] == value for expression, value in zip(self.groupby_expression, values)]
            filter_expression = filter_expressions[0]
            for expression in filter_expressions[1:]:
                filter_expression = filter_expression & expression
        return self.df[filter_expression]

    def __iter__(self):
        if self.combine:
            self.df.execute()
            values = self.by[0].bin_values
            for i, values in enumerate(zip(*values.flatten())):
                values = tuple(k.as_py() for k in values)
                yield values, self.get_group(i)
        else:
            count_agg = vaex.agg.count()
            counts = self.df._agg(count_agg, self.binners)
            mask = counts > 0
            values2d = np.array([coord[mask] for coord in np.meshgrid(*self._coords1d, indexing='ij')], dtype='O')
            for i in range(values2d.shape[1]):
                values = values2d[:,i]
                dff = self.get_group(values)
                yield tuple(values.tolist()), dff

    def __len__(self):
        count_agg = vaex.agg.count()
        counts = self.df._agg(count_agg, self.binners)
        mask = counts > 0
        return mask.sum()

    def _extract_center(self, array):
        # take out the edges
        for i, by in enumerate(self.by):
            array = by.extract_center(i, array)
        return array


class BinBy(GroupByBase):
    """Implementation of the binning and aggregation of data, see :method:`binby`."""

    def __init__(self, df, by, sort=False, ascending=True, progress=None, copy=True):
        super(BinBy, self).__init__(df, by, sort=sort, ascending=ascending, progress=progress, copy=copy)

    def agg(self, actions, merge=False, delay=False, progress=None):
        progressbar_agg = vaex.progress.tree(progress)
        import xarray as xr

        @vaex.delayed
        def process(arrays):
            # take out the edges
            arrays = {key: self._extract_center(value) for key, value in arrays.items()}
            # make sure we respect the sorting
            sorting = tuple(by.sort_indices if by.sort_indices is not None else slice(None) for by in self.by)
            arrays = {key: value[sorting] for key, value in arrays.items()}

            keys = list(arrays.keys())
            key0 = keys[0]
            if not isinstance(actions, collections_abc.Iterable)\
                or isinstance(actions, six.string_types):
                assert len(keys) == 1
                final_array = arrays[key0]
                coords = self._coords1d
                return xr.DataArray(final_array, coords=coords, dims=self.dims)
            else:
                final_array = np.zeros((len(arrays), ) + arrays[key0].shape)
                for i, value in enumerate(arrays.values()):
                    final_array[i] = value
                coords = [list(arrays.keys())] + self._coords1d
                dims = ['statistic'] + self.dims
                return xr.DataArray(final_array, coords=coords, dims=dims)

        result = process(self._agg(actions, progressbar_agg))
        progressbar_agg.exit_on(result)
        if delay:
            return result
        else:
            self.df.execute()
            return result.get()

class GroupBy(GroupByBase):
    """Implementation of the binning and aggregation of data, see :method:`groupby`."""

    def __init__(self, df, by, sort=False, ascending=True, combine=False, expand=True, row_limit=None, copy=True, progress=None):
        super(GroupBy, self).__init__(df, by, sort=sort, ascending=ascending, combine=combine, expand=expand, row_limit=row_limit, copy=copy, progress=progress)

    def agg(self, actions, delay=False, progress=None):
        progressbar = vaex.progress.tree(progress, title="aggregators")
        # TODO: this basically forms a cartesian product, we can do better, use a
        # 'multistage' hashmap
        @vaex.delayed
        def aggregate(promise_by):
            arrays = super(GroupBy, self)._agg(actions, progressbar)
            # we don't want non-existing pairs (e.g. Amsterdam in France does not exist)
            # but also, e.g. GrouperInteger will always expect missing values
            # but they may not aways exist
            counts = self.counts
            # nobody wanted to know count*, but we need it if we included non-existing pairs
            if counts is None and not self.dense:
                count_agg = vaex.agg.count(edges=True)
                counts = self.df._agg(count_agg, self.binners, delay=_USE_DELAY, progress=progressbar)
            arrays = delayed_dict(arrays)
            return counts, arrays

        @vaex.delayed
        def process(args):
            counts, arrays = args
            logger.info(f"aggregated on grid, constructing dataframe...")
            if counts is not None and vaex.array_types.is_scalar(counts):
                counts = np.array([counts])


            for name, array in list(arrays.items()):
                if array is not None and vaex.array_types.is_scalar(array):
                    array = np.array([array])
                    arrays[name] = array

            if counts is not None:
                for name, array in arrays.items():

                    shape1 = vaex.array_types.shape(array)
                    shape2 = vaex.array_types.shape(counts)
                    if shape1 != shape2:
                        raise RuntimeError(f'{array} {name} has shape {shape1} while we expected {shape2}')

            arrays = {key: self._extract_center(value) for key, value in arrays.items()}
            if not self.dense:
                counts = self._extract_center(counts)

            # make sure we respect the sorting
            def sort(ar):
                for i, by in list(enumerate(self.by))[::-1]:
                    sort_indices = by.sort_indices
                    if sort_indices is not None:
                        # if sort_indices come from arrow, it will be uint64
                        # which np.take does not like
                        sort_indices = vaex.array_types.to_numpy(sort_indices)
                        if sort_indices.dtype == np.dtype("uint64"):
                            sort_indices = sort_indices.astype("int64")
                        ar = np.take(ar, sort_indices, axis=i)
                return ar

            arrays = {key: sort(value) for key, value in arrays.items()}

            if self.combine and self.expand and isinstance(self.by[0], GrouperCombined):
                assert len(self.by) == 1
                values = self.by[0].bin_values
                columns = {field.name: ar for field, ar in zip(values.type, values.flatten())}
                for key, value in arrays.items():
                    assert vaex.array_types.ndim(value) == 1
                    columns[key] = value
            else:
                columns = {}
                if self.dense:
                    if len(self.by) == 1:
                        for by in self.by:
                            columns[by.label] = by.bin_values
                    else:
                        array0 = arrays[list(arrays)[0]]
                        # similar to the where, this creates indices like [0, 0, 1, 1, 2, 2], [0, 1, 0, 1, 0, 1]
                        indices = [k.ravel() for k in np.mgrid[[slice(0, n) for n in array0.shape]]]
                        for by, index in zip(self.by, indices):
                            columns[by.label] = vaex.array_types.take(by.bin_values, index)

                else:
                    counts = sort(counts)
                    mask = counts > 0
                    for by, indices in zip(self.by, np.where(mask)):
                        columns[by.label] = by.bin_values.take(indices)
                if self.dense or mask.sum() == mask.size:
                    # if we want all, just take it all
                    # should be faster
                    for key, value in arrays.items():
                        if vaex.array_types.ndim(value) > 1:
                            # only happens for numpy
                            columns[key] = value.ravel()
                        else:
                            columns[key] = value
                else:
                    for key, value in arrays.items():
                        columns[key] = vaex.array_types.filter(value, mask)
            logger.info(f"constructed dataframe")
            dataset_arrays = vaex.dataset.DatasetArrays(columns)
            dataset = DatasetGroupby(dataset_arrays, self.df, self.by_original, actions, combine=self.combine, expand=self.expand, sort=self.sort)
            df_grouped = vaex.from_dataset(dataset)
            return df_grouped
        result = process(delayed_list(aggregate(self._promise_by)))
        progressbar.exit_on(result)
        if delay:
            return result
        else:
            self.df.execute()
            return result.get()

    @docsubst
    def describe(self, expression=None):
        '''Return a dataframe with summary statistics per group for each of the expressions provided.

        Example:

        >>> import vaex
        >>> df = vaex.datasets.titanic()
        >>> df.groupby('pclass').describe('age')
          #    pclass    age_count    age_count_na    age_mean    age_std    age_min    age_max
          0         1          284              39     39.1599    14.5224     0.9167         80
          1         2          261              16     29.5067    13.6125     0.6667         70
          2         3          501             208     24.8164    11.9463     0.1667         74

        Short for:
        >>> df.groupby('pclass').agg({{'age': vaex.agg.describe('age')}}).struct.flatten()
          #    pclass    age_count    age_count_na    age_mean    age_std    age_min    age_max
          0         1          284              39     39.1599    14.5224     0.9167         80
          1         2          261              16     29.5067    13.6125     0.6667         70
          2         3          501             208     24.8164    11.9463     0.1667         74

        :param expression: {expression}
        :rtype: DataFrame
        '''
        if expression is None:
            expression = self.df.get_column_names()
        columns = vaex.utils._ensure_strings_from_expressions(expression)
        columns = vaex.utils._ensure_list(columns)
        aggs = {col: vaex.agg.describe(col) for col in columns}
        return self.agg(aggs).struct.flatten()


@vaex.dataset.register
class DatasetGroupby(vaex.dataset.DatasetDecorator):
    '''Wraps a resulting dataset from a dataframe groupby, so the groupby can be serialized'''
    snake_name = 'groupby'
    def __init__(self, original, df, by, agg, combine, expand, sort):
        assert isinstance(original, vaex.dataset.DatasetArrays)
        super().__init__(original)
        self.df = df
        self.by = by
        self.agg = agg
        self.sort = sort
        self.combine = combine
        self.expand = expand
        self._row_count = self.original.row_count
        self._create_columns()

    def _create_columns(self):
        # we know original is a DatasetArrays
        self._columns = self.original._columns.copy()
        self._ids = self.original._ids.copy()

    @property
    def _fingerprint(self):
        by = self.by
        by = str(by) if not isinstance(by, (list, tuple)) else list(map(str, by))
        id = vaex.cache.fingerprint(self.original.fingerprint, self.df.fingerprint(), by, self.agg, self.combine, self.expand)
        return f'dataset-{self.snake_name}-{id}'

    def chunk_iterator(self, *args, **kwargs):
        yield from self.original.chunk_iterator(*args, **kwargs)

    def hashed(self):
        return type(self)(self.original.hashed(), df=self.df, by=self.by, agg=self.agg, combine=self.combine, expand=self.expand, sort=self.sort)

    def _encode(self, encoding):
        by = self.by
        by = str(by) if not isinstance(by, (list, tuple)) else list(map(str, by))
        spec = {
            'dataframe': encoding.encode('dataframe', self.df),
            'by': by,
            'aggregation': encoding.encode_collection('aggregation', self.agg),
            'combine': self.combine,
            'expand': self.expand,
            'sort': self.sort,
        }
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        df = encoding.decode('dataframe', spec.pop('dataframe'))
        by = spec.pop('by')
        agg = encoding.decode_collection('aggregation', spec.pop('aggregation'))
        sort = spec.pop('sort')
        dfg = df.groupby(by, agg=agg, sort=sort)
        return DatasetGroupby(dfg.dataset.original, df, by, agg, sort=sort, **spec)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['original']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.original = self.df.groupby(self.by, agg=self.agg, sort=self.sort).dataset.original
        # self._create_columns()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return vaex.dataset.DatasetSlicedArrays(self, start=start, end=end)



__author__ = 'breddels'
import pathlib
import logging
from glob import glob as local_glob
import io
import os
import re
import sys
from urllib.parse import parse_qs
import warnings

import pyarrow as pa
import pyarrow.fs

import vaex.file.cache


try:
    from sys import version_info
    if version_info[:2] >= (3, 10):
        from importlib.metadata import entry_points
    else:
        from importlib_metadata import entry_points
except ImportError:
    import pkg_resources
    entry_points = pkg_resources.iter_entry_points

normal_open = open
logger = logging.getLogger("vaex.file")


class FileProxy:
    '''Wraps a file object, giving it a name a dup() method

    The dup is needed since a file is stateful, and needs to be duplicated in threads
    '''
    def __init__(self, file, name, dup, auto_close=True):
        self.file = file
        self.name = name
        self.dup = dup
        self.closed = False
        self.auto_close = auto_close

    def __iter__(self):
        raise NotImplementedError('This is just for looking like a file object to Pandas')

    def write(self, *args):
        return self.file.write(*args)

    def read(self, *args):
        return self.file.read(*args)

    def seek(self, *args):
        return self.file.seek(*args)

    def readinto(self, *args):
        return self.file.readinto(*args)

    def tell(self):
        return self.file.tell()

    def close(self):
        self.closed = True
        return self.file.close()

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        if self.auto_close:
            self.file.close()

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return True

    def closed(self):
        return False

    def flush(self):
        return self.file.flush()


def is_file_object(file):
    return hasattr(file, 'read') and hasattr(file, 'seek')


def file_and_path(file, mode='r', fs_options={}, fs=None):
    if is_file_object(file):
        return file, stringyfy(file)
    else:
        file = open(file, mode=mode, fs_options=fs_options, fs=fs)
        return file, stringyfy(file)


def is_path_like(path):
    try:
        stringyfy(path)
        return True
    except ValueError:
        return False


def stringyfy(path):
    """Get string from path like object of file like object

    >>> import sys, pytest
    >>> if sys.platform.startswith('win'):
    ...     pytest.skip('this doctest does not work on Windows')
    ...
    >>> stringyfy('/tmp/test')
    '/tmp/test'
    >>> from pathlib import Path
    >>> stringyfy(Path('/tmp/test'))
    '/tmp/test'
    """
    try:
        # Pathlib support
        path = path.__fspath__()
    except AttributeError:
        pass
    if hasattr(path, 'name'):  # passed in a file
        path = path.name
    if isinstance(path, str):
        return path
    raise ValueError(f'Cannot convert {path} to a path')


def split_scheme(path):
    path = stringyfy(path)
    if '://' in path:
        scheme, path = path[:path.index('://')], path[path.index('://')+3:]
    else:
        scheme = None
    return scheme, path


def memory_mappable(path):
    path = stringyfy(path)
    scheme, _ = split_scheme(path)
    return scheme is None


def split_options(path, fs_options={}):
    if isinstance(path, list):
        paths = []
        previous_options = None
        for path in path:
            path, options = split_options(path, fs_options)
            if previous_options is not None:
                if previous_options != options:
                    raise ValueError(f'Inconsistent set of fs_options given: {previous_options} {options}')
            else:
                previous_options = options
            paths.append(path)
        return paths, previous_options
    path = stringyfy(path)
    match = re.match(r'(.*?)\?((&?[^=&?]+=[^=&?]+)+)', path)
    if match:
        naked_path, query = match.groups()[:2]
    else:
        naked_path = path
        query = ''
    options = fs_options.copy()
    options.update({key: values[0] for key, values in parse_qs(query).items()})
    return naked_path, options


def split_ext(path, fs_options={}):
    path, fs_options = split_options(path, fs_options=fs_options)
    base, ext = os.path.splitext(path)
    return base, ext, fs_options


def exists(path, fs_options={}, fs=None):
    """Checks if file exists.

    >>> vaex.file.exists('/you/do/not')
    False

    >>> vaex.file.exists('s3://vaex/taxi/nyc_taxi_2015_mini.parquet', fs_options={'anon': True})
    True
    """
    fs, path = parse(path, fs_options=fs_options, fs=fs)
    if fs is None:
        return os.path.exists(path)
    else:
        return fs.get_file_info([path])[0].type != pa.fs.FileType.NotFound


def _get_scheme_handler(path):
    scheme, _ = split_scheme(path)
    for entry in entry_points(group='vaex.file.scheme'):
        if entry.name == scheme:
            return entry.load()
    raise ValueError(f'Do not know how to open {path}, no handler for {scheme} is known')


def remove(path):
    scheme, path = split_scheme(path)
    if scheme:
        raise ValueError('Cannot delete non-local files yet')
    os.remove(path)


def parse(path, fs_options={}, fs=None, for_arrow=False):
    if fs is not None:
        if fs_options:
            warnings.warn(f'Passed fs_options while fs was specified, {fs_options} are ignored')
        if 'fsspec' in sys.modules:
            import fsspec
            if isinstance(fs, fsspec.AbstractFileSystem):
                fs = pa.fs.FSSpecHandler(fs)
            if for_arrow:
                fs = pyarrow.fs.PyFileSystem(fs)
        return fs, path
    if isinstance(path, (list, tuple)):
        scheme, _ = split_scheme(path[0])
    else:
        scheme, _ = split_scheme(path)
    if not scheme:
        return None, path
    if isinstance(path, (list, tuple)):
        module = _get_scheme_handler(path[0])
        return module.parse(path[0], fs_options, for_arrow=for_arrow)[0], path
    else:
        module = _get_scheme_handler(path)
        return module.parse(path, fs_options, for_arrow=for_arrow)


def create_dir(path, fs_options, fs=None):
    fs, path = parse(path, fs_options=fs_options, fs=fs)
    if fs is None:
        fs = pa.fs.LocalFileSystem()
    fs.create_dir(path, recursive=True)


def fingerprint(path, fs_options={}, fs=None):
    """Deterministic fingerprint for a file, useful in combination with dask or detecting file changes.

    Based on mtime (modification time), file size, and the path. May lead to
    false negative if the path changes, but not the content.

    >>> fingerprint('/data/taxi.parquet')  # doctest: +SKIP
    '0171ec50cb2cf71b8e4f813212063a19'

    >>> fingerprint('s3://vaex/taxi/nyc_taxi_2015_mini.parquet', fs_options={'anon': True})  # doctest: +SKIP
    '7c962e2d8c21b6a3681afb682d3bf91b'
    """
    fs, path = parse(path, fs_options, fs=fs)
    path = stringyfy(path)
    if fs is None:
        mtime = os.path.getmtime(path)
        size = os.path.getsize(path)
    else:
        info = fs.get_file_info([path])[0]
        mtime = info.mtime_ns
        size = info.size
    import vaex.cache
    return vaex.cache.fingerprint(('file', (path, mtime, size)))


def size(path, fs_options={}, fs=None):
    """Gives the file size in bytes

    >>> size(os.path.expanduser('~/.vaex/data/helmi-dezeeuw-2000-FeH-v2.hdf5'))  # doctest: +SKIP
    135323168

    >>> size('s3://vaex/taxi/nyc_taxi_2015_mini.parquet', fs_options={'anon': True})
    9820562
    """
    fs, path = parse(path, fs_options, fs=fs)
    path = stringyfy(path)
    if fs is None:
        return os.path.getsize(path)
    else:
        info = fs.get_file_info([path])[0]
        return info.size

def open(path, mode='rb', fs_options={}, fs=None, for_arrow=False, mmap=False, encoding="utf8"):
    '''Return a file like object, also accepts cloud based paths.

    If path is a file-like object already, it will be wrapped in a new file like object that will not
    close the original file.
    This makes is easier to write code like:

        def myfync(f_or_path):
            with vaex.file.open(f_or_path) as f:
                f.write(...)

    Without closing the file when called with an open file.
    '''
    if is_file_object(path):
        if hasattr(path, 'name'):
            name = path.name
        else:
            name = 'unkown name'
        def dup():
            raise RuntimeError('Cannot duplicate this file handle')
        return FileProxy(path, name=name, dup=dup, auto_close=False)
    fs, path = parse(path, fs_options=fs_options, fs=fs, for_arrow=for_arrow)
    if fs is None:
        path = stringyfy(path)
        if for_arrow:
            if fs_options:
                raise ValueError(f'fs_options not supported for local files. You passed: {repr(fs_options)}.')
            if mmap:
                return pa.memory_map(path, mode)
            else:
                return pa.OSFile(path, mode)
        else:
            if 'b' not in mode:
                return normal_open(path, mode, encoding=encoding)
            else:
                return normal_open(path, mode)
    if mode == 'rb':
        def create():
            return fs.open_input_file(path)
    elif mode == "r":
        def create():
            fa = fs.open_input_file(path)
            fp = FileProxy(fa, path, lambda: fs.open_input_file(path))
            return io.TextIOWrapper(fp, encoding=encoding)
    elif mode == 'wb':
        def create():
            return _make_argument_optional(fs.open_output_stream, metadata=None)(path)
    elif mode == "w":
        def create():
            fa =  _make_argument_optional(fs.open_output_stream, metadata=None)(path)
            fp = FileProxy(fa, path, lambda: fs.open_output_stream(path))
            return io.TextIOWrapper(fa, encoding=encoding)
    else:
        raise ValueError(f'Only mode=rb/bw/r/w are supported, not {mode}')
    return FileProxy(create(), path, create)


def _make_argument_optional(f, **defaults):
    # workaround for https://issues.apache.org/jira/browse/ARROW-13546
    # makes f act as if arguments have default values (or ignore when the argument does not exist)
    import inspect
    import functools
    sig = inspect.signature(f)
    params = sig.parameters
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        kwargs = kwargs.copy()
        for name, value in defaults.items():
            if name in params and name not in kwargs:
                kwargs[name] = value
        try:
            return f(*args, **kwargs)
        except TypeError:
            # fallback, just pass all args
            return f(*args, **kwargs, **defaults)

    return wrapper


def dup(file):
    """Duplicate a file like object, s3 or cached file supported"""
    if isinstance(file, (vaex.file.cache.CachedFile, FileProxy)):
        return file.dup()
    else:
        return normal_open(file.name, file.mode)

def glob(path, fs_options={}, fs=None):
    if fs:
        raise ValueError('globbing with custom fs not supported yet, please open an issue.')
    scheme, _ = split_scheme(path)
    if not scheme:
        return local_glob(path)
    module = _get_scheme_handler(path)
    return module.glob(path, fs_options)


def ext(path):
    path = stringyfy(path)
    path, options = split_options(path)
    return os.path.splitext(path)[1]



from __future__ import print_function
import os
import sys
import time


class ProgressBarBase(object):
    def __init__(self, min_value, max_value, title='vaex', format="%(percentage) 6.2f%% %(timeinfo)s cpu: %(cpu_usage)d%%"):
        if title is None:
            title = "vaex"
        self.min_value = min_value
        self.max_value = max_value
        self.format = format
        self.title = title
        self.value = self.min_value
        self.fraction = 0
        self.prevfraction = 0
        self.status = None

    def set_passes(self, passes):
        pass

    def start(self):
        pass

    def exit(self):
        pass

    def set_status(self, status):
        self.status = status

    def update_fraction(self):
        if (self.max_value-self.min_value) == 0:
            self.fraction = 1.
        else:
            self.fraction = (float(self.value)-self.min_value)/(self.max_value-self.min_value)

    def info(self):
        value = self.value
        if value == 0:
            self.prevtime = time.time()
            self.starttime = self.prevtime
            self.utime0, self.stime0 = os.times()[:2]
            self.walltime0 = time.time()
        currenttime = time.time()
        percentage = self.fraction * 100
        self.utime, self.stime = os.times()[:2]
        self.walltime  = time.time()
        cpu_time_delta = self.stime - self.stime0 + self.utime - self.utime0
        wall_time_delta = self.walltime - self.walltime0
        if wall_time_delta == 0:
            cpu_usage = 0  # it's formally infinity
        else:
            cpu_usage = cpu_time_delta / wall_time_delta * 100

        if (self.fraction > 0) and ((self.fraction-self.prevfraction) > 0):
            if self.fraction == 1:
                elapsedtime = (currenttime-self.starttime)
                seconds = elapsedtime
                minutes = seconds/60.
                hours = minutes/60.
                timeinfo = "elapsed time  : % 8.2fs = % 4.1fm = % 2.1fh" % (seconds, minutes, hours)
            else:
                #estimatedtime = (currenttime-self.starttime)/(self.fraction) * (1-self.fraction)
                #estimatedtime = (currenttime-self.prevtime)/(self.fraction-self.prevself.fraction) * (1-self.fraction)
                estimatedtime = (currenttime-self.prevtime) / (self.fraction) * (1-self.fraction)
                seconds = estimatedtime
                minutes = seconds/60.
                hours = minutes/60.
                timeinfo = "estimated time: % 8.2fs = % 4.1fm = % 2.1fh" % (seconds, minutes, hours)
        else:
            timeinfo = "estimated time: unknown                "
        return {"percentage":percentage, "timeinfo":timeinfo, "cpu_usage": cpu_usage, "title": self.title}

    def __repr__(self):
        output = ''
        self.update_fraction()
        return self.format % self.info()


class ProgressBar(ProgressBarBase):
    """
        Implementation of progress bar on a console

        Usage::

            progressbar = ProgressBar(0, 100)
            progressbar.update(0)
            ....
            progressbar.update(100)

        By default, the progress bar writes to stderr, so it doesn't clutter up log files when piping stdout
    """
    def __init__(self, min_value, max_value, title="vaex", format="%(percentage) 6.2f%% %(timeinfo)s", width=40, barchar="#", emptychar="-", output=sys.stdout):
        """		
            :param min_value: minimum value for update(..)
            :param format: format specifier for the output
            :param width: width of the progress bar's (excluding extra text)
            :param barchar: character used to print the bar
            :param output: where to write the output to
        """
        super(ProgressBar, self).__init__(min_value, max_value, format=format, title=title)
        self.width = width
        self.barchar = barchar
        self.emptychar = emptychar
        self.output = output

    def update(self, value):
        self.value = value
        print(repr(self), file=self.output, end=' ')
        self.output.flush()

    def finish(self):
        if self.value != self.max_value:
            self.value = self.max_value
            print(repr(self), file=self.output, end=' ')
        self.output.flush()

    def __repr__(self):
        output = ''
        if self.status:
            bar = f"[{self.status}".ljust(self.width-1, ' ') + ']'
        else:
            self.update_fraction()
            count = int(round(self.fraction * self.width))
            space = self.width - count
            bar = self.title + " [" + (self.barchar * count) + (self.emptychar * space) + "]"
        output = "\r" + bar + super(ProgressBar, self).__repr__()
        if self.fraction == 1:
            output += "\n" # last time print a newline char
        return output


class ProgressBarWidget(ProgressBarBase):
    def __init__(self, min_value, max_value, title=None):
        super(ProgressBarWidget, self).__init__(min_value, max_value)
        import ipywidgets as widgets
        from IPython.display import display
        self.bar = widgets.FloatProgress(min=self.min_value, max=self.max_value)
        self.text = widgets.Label(value='In progress...')
        self.widget = widgets.HBox([self.bar, self.text])
        # self.widget.description = repr(self)
        display(self.widget)

    def __call__(self, value):
        self.value = value
        self.bar.value = value
        self.text.value = repr(self)

    def update(self, value):
        self(value)

    def finish(self):
        self(self.max_value)

import rich.progress
from rich.text import Text


class TimeElapsedColumn(rich.progress.ProgressColumn):
    """Renders time elapsed."""

    def render(self, task: "Task") -> Text:
        """Show time remaining."""
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text("-:--:--", style="progress.elapsed")
        from datetime import timedelta
        delta = timedelta(seconds=(elapsed))
        time = str(delta)[:-4]
        if time.startswith('0:00:'):
            time = time[5:]
        time = time + 's'
        passes = task.fields.get('passes')
        if passes is not None:
            time += f'[{passes}]'
        else:
            time += '   '
        return Text(time, style="progress.elapsed")


class ProgressBarRich(ProgressBarBase):
    def __init__(self, min_value, max_value, title=None, progress=None, indent=0, parent=None):
        super(ProgressBarRich, self).__init__(min_value, max_value, title=title)
        import rich.progress
        import rich.table
        import rich.tree
        self.console = rich.console.Console(record=True)
        self.parent = parent
        if progress is None:
            self.progress = rich.progress.Progress(
                rich.progress.SpinnerColumn(),
                rich.progress.TextColumn("[progress.description]{task.description}"),
                rich.progress.BarColumn(),
                rich.progress.TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                # rich.progress.TimeRemainingColumn(),
                TimeElapsedColumn(),
                rich.progress.TextColumn("[red]{task.fields[status]}"),
                console=self.console,
                transient=False,
                expand=False,
            )
        else:
            self.progress = progress
        if parent is None:
            self.node = rich.tree.Tree(self.progress)
            from rich.live import Live
            self.live = Live(self.node, refresh_per_second=5, console=self.console)
        else:
            self.node = parent.add(self.progress)
        # we do 1000 discrete steps
        self.steps = 0
        self.indent = indent

        padding = max(0, 45- (self.indent * 4) - len(self.title))
        self.passes = None
        self.task = self.progress.add_task(f"[red]{self.title}" + (" " * padding), total=1000, start=False, status=self.status or '', passes=self.passes)
        self.started = False
        self.subtasks = []

    def add_child(self, parent, task, title):
        return ProgressBarRich(self.min_value, self.max_value, title, indent=self.indent+1, parent=self.node)

    def __call__(self, value):
        if not self.started:
            self.progress.start_task(self.task)
        if value > self.value:
            steps = int(value * 1000)
            delta = steps - self.steps
            if delta > 0:
                self.progress.update(self.task, advance=delta, refresh=False, passes=self.passes)
            else:
                start_time = self.progress.tasks[0].start_time
                self.progress.reset(self.task, completed=steps, refresh=False, status=self.status or '')
                self.progress.tasks[0].start_time = start_time
            self.steps = steps
        self.value = value

    def update(self, value):
        self(value)

    def finish(self):
        self(self.max_value)
        if self.parent is None:
            self.live.refresh()

    def start(self):
        if self.parent is None and not self.live.is_started:
            self.live.refresh()
            self.live.start()

    def exit(self):
        if self.parent is None:
            self.live.stop()

    def set_status(self, status):
        self.status = status
        self.progress.update(self.task, status=self.status)

    def set_passes(self, passes):
        self.passes = passes



from __future__ import print_function
__author__ = 'maartenbreddels'
import vaex
import vaex.utils
import logging
import astropy.units
import sys
import yaml

logger = logging.getLogger("vaex.meta")


def main(argv):
    import argparse
    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")

    subparsers = parser.add_subparsers(help='type of subtask', dest="task")

    parser_export = subparsers.add_parser('export', help='read meta info')
    parser_export.add_argument("input", help="input dataset")
    parser_export.add_argument('output', help='output file (.yaml or .json)')
    parser_export.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")
    parser_export.add_argument('--all', dest="all", action='store_true', default=False, help="Also export missing values (useful for having a template)")

    parser_import = subparsers.add_parser('import', help='read meta info')
    parser_import.add_argument('input', help='input meta file (.yaml or .json)')
    parser_import.add_argument("output", help="output dataset")
    parser_import.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")
    parser_import.add_argument('--overwrite', help="overwrite existing entries", default=False, action='store_true')
    parser_import.add_argument('--description', help="overwrite description", default=None)

    args = parser.parse_args(argv[1:])

    verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
    logging.getLogger("vaex").setLevel(verbosity[min(3, args.verbose)])

    if args.task == "export":
        ds = vaex.open(args.input)
        column_names = ds.get_column_names(strings=True, virtual=True)
        if args.all:
            output_data = dict(description=ds.description,
                               descriptions={name: ds.descriptions.get(name, "") for name in column_names},
                               ucds={name: ds.ucds.get(name, "") for name in column_names},
                               units={name: str(ds.units.get(name, "")) for name in column_names},  # {name:str(unit) for name, unit in ds.units.items()},
                               )
        else:
            output_data = dict(description=ds.description,
                               descriptions=ds.descriptions,
                               ucds=ds.ucds,
                               units={name: str(unit) for name, unit in ds.units.items()},
                               )
        if args.output == "-":
            yaml.safe_dump(output_data, sys.stdout, default_flow_style=False)  # , encoding='utf-8',  allow_unicode=True)
        else:
            vaex.utils.write_json_or_yaml(args.output, output_data)
            print("wrote %s" % args.output)
    if args.task == "import":
        if args.input == "-":
            data = yaml.load(sys.stdin)
        else:
            data = vaex.utils.read_json_or_yaml(args.input)

        ds = vaex.open(args.output)

        units = data["units"]
        ucds = data["ucds"]
        descriptions = data["descriptions"]
        if args.description:
            ds.description = args.description
        else:
            if ds.description is None or args.overwrite:
                ds.description = data["description"]
        for column_name in ds.get_column_names(strings=True):
            if column_name not in descriptions:
                print(column_name, 'missing description')
            else:
                print('>>>', column_name, descriptions[column_name])
            if (args.overwrite or column_name not in ds.units) and column_name in units:
                ds.units[column_name] = astropy.units.Unit(units[column_name])
            if (args.overwrite or column_name not in ds.ucds) and column_name in ucds:
                ds.ucds[column_name] = ucds[column_name]
            if (args.overwrite or column_name not in ds.descriptions) and column_name in descriptions:
                ds.descriptions[column_name] = descriptions[column_name]
        ds.write_meta()
        print("updated meta data in %s" % args.output)


if __name__ == "__main__":
    main(sys.argv)



import vaex.utils
import vaex
import os
import numpy as np
# data_dir = "/tmp/vaex/data"
data_dir = vaex.settings.main.data.path

try:
    from urllib import urlretrieve  # py2
except ImportError:
    from urllib.request import urlretrieve  # py3


def _url_to_filename(url, replace_ext=None, subdir=None):
    if subdir:
        filename = os.path.join(data_dir, subdir, url.split("/")[-1])
    else:
        filename = os.path.join(data_dir, url.split("/")[-1])
    if replace_ext:
        dot_index = filename.rfind(".")
        filename = filename[:dot_index] + replace_ext
    return filename


class Hdf5Download(object):
    def __init__(self, url):
        self.url = url
        self.url_list = [self.url]

    @property
    def filename(self):
        os.makedirs(data_dir, exist_ok=True)
        return os.path.join(data_dir, _url_to_filename(self.url))

    def download(self, force=False):
        if not os.path.exists(self.filename) or force:
            print("Downloading %s to %s" % (self.url, self.filename))
            code = os.system(self.wget_command(0))
            if not os.path.exists(self.filename):
                print("wget failed, using urlretrieve")
                self.download_urlretrieve()

    def download_urlretrieve(self):
        urlretrieve(self.url, self.filename)

    def fetch(self, force_download=False):
        self.download(force=force_download)
        return vaex.open(self.filename, convert=not self.filename.endswith('.hdf5'), progress=True)

    def wget_command(self, i):
        assert i == 0
        url = self.url_list[i]
        return "wget --progress=bar:force -c -P %s %s" % (data_dir, url)


def iris():
    '''Load and return the iris dataset (classification).

    The iris dataset is a classic and very easy multi-class classification dataset.

    =================   ==============
    Classes                          3
    Samples per class               50
    Samples total                  150
    Dimensionality                   4
    Features            real, positive
    =================   ==============

    Example:

    >>> import vaex.datasets
    >>> df = vaex.datasets.datasets.iris()
    >>> df.describe()
    '''
    dirname = os.path.dirname(__file__)
    return vaex.open(os.path.join(dirname, 'iris.hdf5'))


def _iris(name, iris_previous, N):
    filename = os.path.join(vaex.settings.data.path, name + '.hdf5')
    if os.path.exists(filename):
        return vaex.open(filename)
    else:
        iris = iris_previous()
        repeat = int(np.ceil(N / len(iris)))
        ds = vaex.concat([iris] * repeat)
        ds.export_hdf5(filename)
        return vaex.open(filename)


def iris_subsample(N, error_percentage=5, ds=None):
    '''Returns the iris set repeated so it include ~1e4 rows'''
    # return _iris_subsample('iris_1e4', iris, int(1e4))
    ds = ds or iris()
    ds_out = None
    repeats = int(np.ceil(N / len(ds)))
    for feature in ds.get_column_names():
        if feature in ['random_index']:
            continue
        data = ds[feature].values
        min, max = ds.minmax(feature)
        error = (max - min)/100*error_percentage
        data_out = np.repeat(data, repeats)
        if feature not in ['class_']:
            data_out += np.random.random(len(data_out)) * error
        if ds_out is None:
            ds_out = vaex.from_arrays(feature=data_out)
        else:
            ds_out.add_column(feature, data_out)
    return ds_out


def iris_1e4():
    '''Returns the iris set repeated so it include ~1e4 rows'''
    return _iris('iris_1e4', iris, int(1e4))


def iris_1e5():
    '''Returns the iris set repeated so it include ~1e5 rows'''
    return _iris('iris_1e5', iris_1e4, int(1e5))


def iris_1e6():
    '''Returns the iris set repeated so it include ~1e6 rows'''
    return _iris('iris_1e6', iris_1e5, int(1e6))


def iris_1e7():
    '''Returns the iris set repeated so it include ~1e7 rows'''
    return _iris('iris_1e7', iris_1e6, int(1e7))


def iris_1e8():
    '''Returns the iris set repeated so it include ~1e8 rows'''
    return _iris('iris_1e8', iris_1e7, int(1e8))


def iris_1e9():
    '''Returns the iris set repeated so it include ~1e8 rows'''
    return _iris('iris_1e9', iris_1e8, int(1e9))


def titanic():
    '''
    Returns the classic Titanic dataset.

    Description of the columns can be found in dataset.description.

    Example:

    >>> import vaex.datasets
    >>> df = vaex.datasets.titanic()
    >>> print(df.description)
    >>> df.describe()
    '''
    dirname = os.path.dirname(__file__)
    return vaex.open(os.path.join(dirname, 'titanic.hdf5'))


def taxi():
    '''One year of NYC Yellow Cab data.

    The original raw data can be downloaded from https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
    '''
    return Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/1.1/yellow_taxi_2012_zones.parquet").fetch()


def helmi_simulation_data(full=False):
    '''Result of an N-body simulation of the accretion of 33 satellite galaxies into a Milky Way dark matter halo.

    Data was greated by Helmi & de Zeeuw 2000.
    The data contains the position (x, y, z), velocitie (vx, vy, vz), the energy (E),
    the angular momentum (L, Lz) and iron content (FeH) of the particles.

    :param bool full: If True, it returns the full output of the simulation, i.e 3.3 million rows. Otherwise only 10% is returned.
    :rtype: DataFrame
    '''
    if full is False:
        return Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/helmi-dezeeuw-2000-FeH-v2-10percent.hdf5").fetch()
    return Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/helmi-dezeeuw-2000-FeH-v2.hdf5").fetch()


def tgas(full=False):
    '''The Tycho-Gaia Astronomical Solution Dataset.

    This dataset combines data of the Hipparcos and Tycho catalogues with the Gaia DR1
    catalogue in order to provide a full astrometric solution for the Gaia data for
    the stars in common.

    :param bool full: If True, it returns the full output of the simulation, i.e 2 million rows. Otherwise only 1% is returned.

    For more information about this datasets visit https://gea.esac.esa.int/archive/documentation/GDR1/Data_processing/chap_cu3tyc/
    '''
    if full:
        return Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/tgas.hdf5").fetch()
    else:
        return Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/tgas_1percent.hdf5").fetch()


# TODO: deprecate in v5
helmi_de_zeeuw = Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/helmi-dezeeuw-2000-FeH-v2.hdf5")



import numpy as np
from vaex.dataframe import DataFrame

import vaex.dataset
from vaex.utils import _ensure_string_from_expression

@vaex.dataset.register
class DatasetJoin(vaex.dataset.DatasetDecorator):
    snake_name = "join"
    _no_serialize = '_columns _ids original _row_count _id _cached_fingerprint'.split()
    def __init__(self, original, left, right, on=None, left_on=None, right_on=None, lprefix='', rprefix='', lsuffix='', rsuffix='', how='left', allow_duplication=False, prime_growth=False, cardinality_other=None, slice_start=None, slice_end=None):
        super().__init__(original)
        self.left = left
        self.right = right
        # trigger length computation, otherwise it will be triggered in .slice, which can lead
        # to nested executions
        len(left)
        len(right)
        # len(left)
        # len(right)
        self.on = on
        self.left_on = left_on
        self.right_on = right_on
        self.lprefix = lprefix
        self.rprefix = rprefix
        self.lsuffix = lsuffix
        self.rsuffix = rsuffix
        self.how = how
        self.allow_duplication = allow_duplication
        self.prime_growth = prime_growth
        self.cardinality_other = cardinality_other
        self.slice_start = slice_start
        self.slice_end = slice_end
        self._row_count = self.original.row_count
        self._create_columns()

    def _create_columns(self):
        # we know original is a DatasetArrays
        self._columns = self.original._columns.copy()
        self._ids = self.original._ids.copy()

    @property
    def _fingerprint(self):
        # kwargs = self.__dict__.copy()
        # for skip in self._no_serialize:
        #     kwargs.pop(skip)
        fp = vaex.cache.fingerprint({
            'left': self.left.fingerprint(),
            'right': self.right.fingerprint(),
            'on': str(self.on) if self.on is not None else None,
            'left_on': str(self.left_on) if self.left_on is not None else None,
            'right_on': str(self.right_on) if self.right_on is not None else None,
            'lprefix': self.lprefix,
            'rprefix': self.rprefix,
            'lsuffix': self.lsuffix,
            'rsuffix': self.rsuffix,
            'how': self.how,
            'allow_duplication': self.allow_duplication,
            'prime_growth': self.prime_growth,
            'cardinality_other': self.cardinality_other,
        })
        return f'dataset-{self.snake_name}-{fp}'

    def chunk_iterator(self, *args, **kwargs):
        yield from self.original.chunk_iterator(*args, **kwargs)

    def hashed(self):
        kwargs = self.__dict__.copy()
        for skip in self._no_serialize:
            kwargs.pop(skip, None)
        return type(self)(self.original.hashed(), **kwargs)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        kwargs = self.__dict__.copy()
        kwargs['slice_start'] = start
        kwargs['slice_end'] = end
        for skip in self._no_serialize:
            kwargs.pop(skip, None)
        left = kwargs.pop('left')
        right = kwargs.pop('right')
        dataset = self.original.slice(start, end)
        return type(self)(dataset, left=left, right=right, **kwargs)

    def _encode(self, encoding):
        # by = self.by
        # by = str(by) if not isinstance(by, (list, tuple)) else list(map(str, by))
        spec = {
            'left': encoding.encode('dataframe', self.left),
            'right': encoding.encode('dataframe', self.right),
            'on': str(self.on) if self.on is not None else None,
            'left_on': str(self.left_on) if self.left_on is not None else None,
            'right_on': str(self.right_on) if self.right_on is not None else None,
            'lprefix': self.lprefix,
            'rprefix': self.rprefix,
            'lsuffix': self.lsuffix,
            'rsuffix': self.rsuffix,
            'how': self.how,
            'allow_duplication': self.allow_duplication,
            'prime_growth': self.prime_growth,
            'cardinality_other': self.cardinality_other,
            'slice_start': self.slice_start,
            'slice_end': self.slice_end,
        }
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        spec = spec.copy()
        left = encoding.decode('dataframe', spec.pop('left'))
        right = encoding.decode('dataframe', spec.pop('right'))
        spec_no_slice = spec.copy()
        del spec_no_slice['slice_start']
        del spec_no_slice['slice_end']
        dfj = left.join(right, **spec_no_slice)
        start = spec['slice_start']
        end = spec['slice_end']
        dataset = dfj.dataset.original
        if start is not None and end is not None:
            dataset = dataset.slice(start, end)
        return DatasetJoin(dataset, left, right, **spec)
        

def join(df, other, on=None, left_on=None, right_on=None, lprefix='', rprefix='', lsuffix='', rsuffix='', how='left', allow_duplication=False, prime_growth=False, cardinality_other=None, inplace=False):
    # implementation of DataFrameLocal.join
    inner = False
    left: DataFrame = df
    right: DataFrame = other
    left_original = left.copy()
    right_original = right.copy()
    rprefix_original, lprefix_original = rprefix, lprefix
    rsuffix_original, lsuffix_original = rsuffix, lsuffix
    right_on_original, left_on_original = right_on, left_on
    if how == 'left':
        pass
    elif how == 'right':
        left, right = right, left
        lprefix, rprefix = rprefix, lprefix
        lsuffix, rsuffix = rsuffix, lsuffix
        left_on, right_on = right_on, left_on
    elif how == 'inner':
        inner = True
    else:
        raise ValueError('join type not supported: {}, only left and right'.format(how))
    left = left if inplace else left.copy()

    on = _ensure_string_from_expression(on)
    left_on = _ensure_string_from_expression(left_on)
    right_on = _ensure_string_from_expression(right_on)
    left_on = left_on or on
    right_on = right_on or on
    for name in right:
        if left_on and (rprefix + name + rsuffix == lprefix + left_on + lsuffix):
            continue  # it's ok when we join on the same column name
        if name in left and rprefix + name + rsuffix == lprefix + name + lsuffix:
            raise ValueError('column name collision: {} exists in both column, and no proper suffix given'
                                .format(name))

    right = right.extract()  # get rid of filters and active_range
    assert left.length_unfiltered() == left.length_original()
    N = left.length_unfiltered()
    N_other = len(right)
    if left_on is None and right_on is None:
        lookup = None
    else:
        df = left
        # we index the right side, this assumes right is smaller in size
        index = right._index(right_on, prime_growth=prime_growth, cardinality=cardinality_other)
        dtype = left.data_type(left_on)
        duplicates_right = index.has_duplicates

        if duplicates_right and not allow_duplication:
            raise ValueError('This join will lead to duplication of rows which is disabled, pass allow_duplication=True')

        # our max value for the lookup table is the row index number, so if we join a small
        # df with say 100 rows, we can do it with a int8
        lookup_dtype = vaex.utils.required_dtype_for_max(len(right))
        # we put in the max value to maximize triggering failures in the case of a bug (we don't want
        # to point to row 0 in case we do, we'd rather crash)
        lookup = np.full(left._length_original, np.iinfo(lookup_dtype).max, dtype=lookup_dtype)
        nthreads = df.executor.thread_pool.nthreads
        lookup_masked = [False] * nthreads  # does the lookup contain masked/-1 values?
        lookup_extra_chunks = []

        from vaex.column import _to_string_sequence
        def map(thread_index, i1, i2, selection_masks, blocks):
            ar = blocks[0]
            if vaex.array_types.is_string_type(dtype):
                previous_ar = ar
                ar = _to_string_sequence(ar)
            if dtype.is_datetime:
                ar = ar.view(np.int64)
            if np.ma.isMaskedArray(ar):
                mask = np.ma.getmaskarray(ar)
                found_masked = index.map_index_masked(ar.data, mask, lookup[i1:i2])
                lookup_masked[thread_index] = lookup_masked[thread_index] or found_masked
                if duplicates_right:
                    extra = index.map_index_duplicates(ar.data, mask, i1)
                    lookup_extra_chunks.append(extra)
            else:
                found_masked = index.map_index(ar, lookup[i1:i2])
                lookup_masked[thread_index] = lookup_masked[thread_index] or found_masked
                if duplicates_right:
                    extra = index.map_index_duplicates(ar, i1)
                    lookup_extra_chunks.append(extra)
        def reduce(a, b):
            pass
        left.map_reduce(map, reduce, [left_on], delay=False, name='fill looking', info=True, to_numpy=False, ignore_filter=True)
        if len(lookup_extra_chunks):
            # if the right has duplicates, we increase the left of left, and the lookup array
            lookup_left = np.concatenate([k[0] for k in lookup_extra_chunks])
            lookup_right = np.concatenate([k[1] for k in lookup_extra_chunks])
            left = left.concat(left.take(lookup_left))
            lookup = np.concatenate([lookup, lookup_right])

        if inner:
            left_mask_matched = lookup != -1  # all the places where we found a match to the right
            lookup = lookup[left_mask_matched]  # filter the lookup table to the right
            left_indices_matched = np.where(left_mask_matched)[0]  # convert mask to indices for the left
            # indices can still refer to filtered rows, so do not drop the filter
            left = left.take(left_indices_matched, filtered=False, dropfilter=False)
    direct_indices_map = {}  # for performance, keeps a cache of two levels of indirection of indices

    def mangle_name(prefix, name, suffix):
        if name.startswith('__'):
            return '__' + prefix + name[2:] + suffix
        else:
            return prefix + name + suffix

    # first, do renaming, so all column names are unique
    right_names = right.get_names(hidden=True)
    left_names = left.get_names(hidden=True)
    for name in right_names:
        if name in left_names:
            # find a unique name across both dataframe, including the new name for the left
            all_names = list(set(right_names + left_names))
            all_names.append(mangle_name(lprefix, name, lsuffix))  # we dont want to steal the left's name
            all_names.remove(name)  # we could even claim the original name
            new_name = mangle_name(rprefix, name, rsuffix)
            # we will not add this column twice when it is the join column
            if new_name != left_on:
                if new_name in all_names:  # it's still not unique
                    new_name = vaex.utils.find_valid_name(new_name, all_names)
                right.rename(name, new_name)
                right_names[right_names.index(name)] = new_name

            # and the same for the left
            all_names = list(set(right_names + left_names))
            all_names.remove(name)
            new_name = mangle_name(lprefix, name, lsuffix)
            if new_name in all_names:  # still not unique
                new_name = vaex.utils.find_valid_name(new_name, all_names)
            left.rename(name, new_name)
            left_names[left_names.index(name)] = new_name

    # now we add columns from the right, to the left
    right_names = right.get_names(hidden=True)
    left_names = left.get_names(hidden=True)
    right_columns = []
    for name in right_names:
        column_name = name
        if name == left_on and name in left_names:
            continue  # skip when it's the join column
        assert name not in left_names
        if name in right.variables:
            left.set_variable(name, right.variables[name])
        elif column_name in right.virtual_columns:
            left.add_virtual_column(name, right.virtual_columns[column_name])
        elif column_name in right.functions:
            if name in left.functions:
                raise NameError(f'Name collision for function {name}')
            left.functions[name] = right.functions[name]
        else:
            right_columns.append(name)
            # we already add the column name here to get the same order
            left.column_names.append(name)
            left._initialize_column(name)
    # merge the two datasets
    if right_columns:
        right_dataset = right.dataset.project(*right_columns)
        if lookup is not None:
            # if lookup is None, we do a row based join
            # and we only need to merge.
            # if we have an array of lookup indices, we 'take' those
            right_dataset = right_dataset.take(lookup, masked=any(lookup_masked))
        dataset = left.dataset.merged(right_dataset)
    else:
        dataset = left.dataset
    # row number etc should not have changed, we only append new columns
    # so no need to reset caches
    left._dataset = DatasetJoin(dataset, left_original, right_original,
        on=on, left_on=left_on_original, right_on=right_on_original,
        lprefix=lprefix_original, rprefix=rprefix_original, lsuffix=lsuffix_original, rsuffix=rsuffix_original,
        how=how, allow_duplication=allow_duplication, prime_growth=prime_growth, cardinality_other=cardinality_other
    )
    return left



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import asyncio
import threading
import multiprocessing
from warnings import warn
import vaex.utils
import logging
import concurrent.futures
import time
import vaex.settings

from .itertools import buffer

logger = logging.getLogger("vaex.multithreading")

main_pool = None
main_io_pool = None


thread_pools = {}


def get_thread_pool(nthreads):
    if nthreads not in thread_pools:
        thread_pools[nthreads] = ThreadPoolIndex(nthreads)
    return thread_pools[nthreads]


def get_main_pool():
    global main_pool
    if main_pool is None:
        main_pool = ThreadPoolIndex()
        thread_pools[main_pool.nthreads] = main_pool
    return main_pool


def get_main_io_pool():
    global main_io_pool
    if main_io_pool is None:
        main_io_pool = concurrent.futures.ThreadPoolExecutor(max_workers=vaex.settings.main.thread_count_io)
    return main_io_pool





class ThreadPoolIndex(concurrent.futures.ThreadPoolExecutor):
    """Thread pools that adds a thread index as first argument to the callback passed to map

    This is useful if you keep a piece of memory (like ndgrid) per array
    """

    def __init__(self, max_workers=None, *args, **kwargs):
        if max_workers is None:
            max_workers = vaex.settings.main.thread_count
        super(ThreadPoolIndex, self).__init__(max_workers, *args, **kwargs)
        self.lock = threading.Lock()
        self.thread_indices = iter(range(1000000))  # enough threads until 2100?
        self.local = threading.local()
        self.nthreads = self._max_workers
        self._debug_sleep = 0

    def map(self, callable, iterator, count, on_error=None, progress=None, cancel=None, unpack=False, use_async=False, **kwargs_extra):
        progress = progress or (lambda x: True)
        cancelled = False

        def wrapped(*args, **kwargs):
            if not cancelled:
                if self.nthreads == 1:
                    self.local.index = 0
                with self.lock:
                    if not hasattr(self.local, 'index'):
                        self.local.index = next(self.thread_indices)
                if unpack:
                    args = args[0]  # it's passed as a tuple.. not sure why
                if self._debug_sleep:
                    # print("SLEEP", self._debug_sleep)
                    time.sleep(self._debug_sleep)
                return callable(self.local.index, *args, **kwargs, **kwargs_extra)
        time_last = time.time() - 100
        min_delta_t = 1. / 10  # max 10 per second
        # we don't want to keep consuming the chunk iterator when we cancel
        chunk_iterator = iterator
        def cancellable_iter():
            for value in chunk_iterator:
                yield value
                if cancelled:
                    break
        if self.nthreads == 1:  # when using 1 thread, it makes debugging easier (better stacktrace)
            if use_async:
                iterator = self._map_async(wrapped, cancellable_iter())
            else:
                iterator = self._map(wrapped, cancellable_iter())
        else:
            if use_async:
                loop = asyncio.get_event_loop()
                iterator = (loop.run_in_executor(self, lambda value=value: wrapped(value)) for value in cancellable_iter())
            else:
                iterator = super(ThreadPoolIndex, self).map(wrapped, cancellable_iter())
        total = 0
        iterator = iter(buffer(iterator, self._max_workers + 3))
        try:
            for value in iterator:
                if use_async:
                    value = yield value
                else:
                    yield value
                if value != None:
                    total += value
                progress_value = (total) / count
                time_now = time.time()
                if progress_value == 1 or (time_now - time_last) > min_delta_t:
                    time_last = time_now
                    if progress(progress_value) is False:
                        cancelled = True
                        cancel()
                        break
        finally:
            if not cancelled:
                cancelled = True
            # consume the rest of the iterators and await them to avoid un-awaited exceptions, which trigger a
            # 'Future exception was never retrieved' printout
            # TODO: since we don't use async any more, I think we can get rid of this
            for value in iterator:
                try:
                    pass
                except:
                    pass

    def _map(self, callable, iterator):
        for i in iterator:
            yield callable(i)

    def _map_async(self, callable, iterator):
        for i in iterator:
            future = asyncio.Future()
            future.set_result(callable(i))
            yield future



import copyreg
import sys
import os

import dask.base
import pyarrow as pa
import numpy as np

import vaex
import vaex.array_types
from vaex.column import _to_string_sequence
import vaex.utils


if vaex.utils.has_c_extension:
    from .superutils import *
    from . import superutils

    ordered_set = tuple([cls for name, cls in vars(superutils).items() if name.startswith('ordered_set')])

    def pickle(x):
        keys = x.key_array()
        return type(x), (keys, x.null_index, x.nan_count, x.null_count, x.fingerprint)
    for cls in ordered_set:
        copyreg.pickle(cls, pickle)


    def create_set_string(keys, *other):
        keys = vaex.column.ColumnStringArrow.from_arrow(keys)
        return ordered_set_string(keys.string_sequence, *other)

    def pickle_set_string(x):
        keys = x.key_array()
        keys = pa.array(keys.to_numpy(), type=pa.large_utf8())
        return create_set_string, (keys, x.null_index, x.nan_count, x.null_count, x.fingerprint)
    copyreg.pickle(ordered_set_string, pickle_set_string)

    for ordered_set_cls in ordered_set:
        @dask.base.normalize_token.register(ordered_set_cls)
        def normalize(obj):
            if not obj.fingerprint:
                raise RuntimeError('No fingerprint present in set')
            return obj.fingerprint


def counter_type_from_dtype(dtype, transient=True):
    return vaex.utils.find_type_from_dtype(vaex.hash, 'counter_', dtype, transient=transient, support_non_native=False)


def ordered_set_type_from_dtype(dtype, transient=True):
    name = 'ordered_set_'
    return vaex.utils.find_type_from_dtype(vaex.hash, name, dtype, transient=transient, support_non_native=False)


def index_type_from_dtype(dtype, transient=True, prime_growth=False):
    name = 'index_hash_'
    if prime_growth:
        name += "_prime_growth"
    return vaex.utils.find_type_from_dtype(vaex.hash, name, dtype, transient=transient, support_non_native=False)


@vaex.encoding.register("hash-map-unique")
class HashMapUnique:
    '''HashMap that maps keys to unique integers'''
    def __init__(self, dtype, nmaps=1, limit=None, _internal=None):
        self.dtype = dtype
        self.dtype_item = self.dtype
        limit = -1 if limit is None else limit
        if _internal is None:
            cls = ordered_set_type_from_dtype(dtype)
            self._internal = cls(nmaps, limit)
        else:
            self._internal = _internal

    def flatten(self):
        if self.dtype == object:
            return self # already flat
        keys = self._internal.key_array()
        map = type(self._internal)(keys, self.null_index, self.nan_count, self.null_count, self.fingerprint)
        return HashMapUnique(self.dtype, _internal=map)

    @staticmethod
    def encode(encoding, obj):
        keys = obj._internal.key_array()
        if isinstance(keys, (vaex.strings.StringList32, vaex.strings.StringList64)):
            keys = vaex.strings.to_arrow(keys)
        keys = encoding.encode('array', keys)
        clsname = obj._internal.__class__.__name__
        return {
            'class': clsname,
            'dtype': encoding.encode('dtype', vaex.dtype(obj.dtype)),
            'data': {
                'keys': keys,
                'null_index': obj.null_index,
                'nan_count': obj.nan_count,
                'missing_count': obj.null_count,
                'fingerprint': obj.fingerprint,
            }
        }

    @staticmethod
    def decode(encoding, obj_spec):
        clsname = obj_spec['class']
        cls = getattr(vaex.hash, clsname)
        keys = encoding.decode('array', obj_spec['data']['keys'])
        dtype = vaex.dtype_of(keys)
        if dtype.is_string:
            keys = vaex.strings.to_string_sequence(keys)
        _hash_map_internal = cls(keys, obj_spec['data']['null_index'], obj_spec['data']['nan_count'], obj_spec['data']['missing_count'], obj_spec['data']['fingerprint'])
        dtype = encoding.decode('dtype', obj_spec['dtype'])
        return vaex.hash.HashMapUnique(dtype, _internal=_hash_map_internal)

    @classmethod
    def from_keys(cls, keys, dtype=None, fingerprint=''):
        keys = vaex.array_types.convert(keys, 'numpy-arrow')
        dtype = vaex.dtype_of(keys) if dtype is None else dtype
        if dtype == float:
            nancount = np.isnan(keys).sum()
        else:
            nancount = 0
        null_count = 0
        null_index = -1
        if np.ma.isMaskedArray(keys):
            null_count = keys.mask.sum()
            if null_count == 0:
                keys = keys.data
            elif null_count == 1:
                null_index = np.where(keys.mask == 1)[0]
                keys = keys.data
            else:
                raise ValueError('key arrays contained more than 1 null value')
        set_type = vaex.hash.ordered_set_type_from_dtype(dtype)
        if dtype.is_string:
            # TODO: support 32 bit in hashmap
            keys = keys.cast(pa.large_string())
            values = vaex.column.ColumnStringArrow.from_arrow(keys)
            string_sequence = values.string_sequence
            mask = string_sequence.mask()
            if mask is not None:
                null_count = mask.sum()
                if null_count == 0:
                    pass  # fine
                elif null_count == 1:
                    null_index = np.where(mask == 1)[0]
                else:
                    raise ValueError('key arrays contained more than 1 null value')
            hash_map_unique_internal = set_type(string_sequence, null_index, nancount, null_count, fingerprint)
        else:
            hash_map_unique_internal = set_type(keys, null_index, nancount, null_count, fingerprint)
        return HashMapUnique(dtype, _internal=hash_map_unique_internal)

    def add(self, ar, return_inverse=False):
        if self.dtype_item.is_string:
            ar = _to_string_sequence(ar)
        else:
            ar = vaex.array_types.to_numpy(ar)
            if ar.strides != (ar.itemsize,):
                ar = ar.copy()

        chunk_size = 1024*1024
        if np.ma.isMaskedArray(ar):
            mask = np.ma.getmaskarray(ar)
            if return_inverse:
                return self._internal.update(ar, mask, -1, chunk_size=chunk_size, bucket_size=chunk_size*4, return_values=return_inverse)
            else:
                self._internal.update(ar, mask,  -1, chunk_size=chunk_size, bucket_size=chunk_size*4)
        else:
            if return_inverse:
                return self._internal.update(ar, -1, chunk_size=chunk_size, bucket_size=chunk_size*4, return_values=return_inverse)
            else:
                self._internal.update(ar, -1, chunk_size=chunk_size, bucket_size=chunk_size*4)
    
    def __len__(self):
        return len(self._internal)

    def merge(self, others):
        self._internal.merge(others)

    def keys(self, mask=True):
        ar = self._internal.key_array()
        if self.dtype_item.is_datetime or self.dtype_item.is_timedelta:
            ar = ar.view(self.dtype_item.numpy)
        if isinstance(ar, vaex.superstrings.StringList64):
            # TODO: find out why this more efficient path does not work
            ar = vaex.column.ColumnStringArrow.from_string_sequence(ar)
            ar = pa.array(ar)
        if mask and self.has_null and (self.dtype_item.is_primitive or self.dtype_item.is_datetime):
            mask = np.zeros(shape=ar.shape, dtype="?")
            mask[self.null_index] = 1
            ar = np.ma.array(ar, mask=mask)
        return ar

    def map(self, keys, check_missing=False):
        '''Map key values to unique integers'''
        from vaex.column import _to_string_sequence

        if not isinstance(keys, vaex.array_types.supported_array_types) or self.dtype == str:
            # sometimes the dtype can be object, but seen as an string array
            keys = _to_string_sequence(keys)
        else:
            keys = vaex.array_types.to_numpy(keys)
        indices = self._internal.map_ordinal(keys)
        if np.ma.isMaskedArray(keys):
            indices[keys.mask] = self.null_index
        if check_missing:
            indices = np.ma.array(indices, mask=indices==-1)
        return indices

    def isin(self, values):
        if vaex.column._is_stringy(values) or self.dtype_item == str:
            values = vaex.column._to_string_column(values)
            values = _to_string_sequence(values)
            # return x.string_sequence.isin(values)
            return self._internal.isin(values)
        else:
            if np.ma.isMaskedArray(values):
                isin = self._internal.isin(values.data)
                isin[values.mask] = False
                return isin
            else:
                return self._internal.isin(values)

    @property
    def has_null(self):
        return self._internal.has_null

    @property
    def has_nan(self):
        return self._internal.has_nan

    @property
    def null_index(self):
        return self._internal.null_index

    @property
    def nan_index(self):
        return self._internal.nan_index

    @property
    def fingerprint(self):
        return self._internal.fingerprint

    @property
    def nan_count(self):
        return self._internal.nan_count

    @property
    def null_count(self):
        return self._internal.null_count

    def sorted(self, keys=None, ascending=True, indices=None, return_keys=False):
        keys = self.keys() if keys is None else keys
        keys = vaex.array_types.to_arrow(keys)
        indices = pa.compute.sort_indices(keys, sort_keys=[('x', "ascending" if ascending else "descending")]) if indices is None else indices
        # arrow sorts with null last
        null_index = -1 if not self.has_null else len(keys)-1
        keys = pa.compute.take(keys, indices)
        fingerprint = self._internal.fingerprint + "-sorted"
        if self.dtype_item.is_string:
            # TODO: supported 32 bit in hashmap
            keys = keys.cast(pa.large_string())
            ar = vaex.column.ColumnStringArrow.from_arrow(keys)
            string_sequence = ar.string_sequence

            set = type(self._internal)(string_sequence, null_index, self._internal.nan_count, self._internal.null_count, fingerprint)
        else:
            set = type(self._internal)(keys, null_index, self._internal.nan_count, self._internal.null_count, fingerprint)
        hash_map_sorted = HashMapUnique(self.dtype, _internal=set)
        if return_keys:
            return hash_map_sorted, keys
        else:
            return hash_map_sorted

    def limit(self, limit):
        fingerprint = self.fingerprint + f"-limit-{limit}"
        keys = self.keys(mask=False)
        keys = keys[:limit]
        null_index = self.null_index
        null_count = 1
        nan_count = 0
        if null_index >= limit:
            null_index = -1
            null_count = 0
        if self.dtype_item == float:
            nan_count = np.isnan(keys).sum()
        if self.dtype_item.is_string:
            bin_values = vaex.column.ColumnStringArrow.from_arrow(keys)
            string_sequence = bin_values.string_sequence
            hash_map_unique = type(self._internal)(string_sequence, null_index, nan_count, null_count, fingerprint)
        else:
            hash_map_unique = type(self._internal)(keys, null_index, nan_count, null_count, fingerprint)
        hash_map_unique.fingerprint = fingerprint
        return HashMapUnique(self.dtype, _internal=hash_map_unique)

    def __sizeof__(self) -> int:
        return sys.getsizeof(self._internal)


@dask.base.normalize_token.register(HashMapUnique)
def normalize(obj):
    if not obj.fingerprint:
        raise RuntimeError('No fingerprint present in HashMapUnique')
    return obj.fingerprint



"""Sets up logging for vaex.

See `configuration of logging <conf.html#logging>`_ how to configure logging.

"""
import os
import vaex.settings
import logging

logger = logging.getLogger('vaex')
log_handler : logging.Handler = None


def set_log_level(loggers=["vaex"], level=logging.DEBUG):
    """set log level to debug"""
    for logger in loggers:
        logging.getLogger(logger).setLevel(level)


def set_log_level_debug(loggers=["vaex"]):
    """set log level to debug"""
    set_log_level(loggers, logging.DEBUG)


def set_log_level_info(loggers=["vaex"]):
    """set log level to info"""
    set_log_level(loggers, logging.INFO)


def set_log_level_warning(loggers=["vaex"]):
    """set log level to warning"""
    set_log_level(loggers, logging.WARNING)


def set_log_level_error(loggers=["vaex"]):
    """set log level to exception/error"""
    set_log_level(loggers, logging.ERROR)


def remove_handler():
    """Disabled logging, remove default hander and add null handler"""
    logging.getLogger('vaex').removeHandler(log_handler)
    logging.getLogger('vaex').addHandler(logging.NullHandler())

def reset():
    '''Reset configuration of logging (i.e. remove the default handler)'''
    logging.getLogger('vaex').removeHandler(log_handler)


def _set_log_level(conf, level):
    if conf:
        if conf.startswith('vaex'):
            set_log_level(conf.split(","), level=level)
        else:
            set_log_level(level=level)

def setup():
    """Setup logging based on the configuration in ``vaex.settings``

    This function is automatically called when importing vaex. If settings are changed, call :func:`reset` and this function again
    to re-apply the settings.
    """
    global log_handler

    if vaex.settings.main.logging.setup:
        logger.setLevel(logging.DEBUG)

        # create console handler and accept all loglevels
        if vaex.settings.main.logging.rich:
            from rich.logging import RichHandler
            log_handler = RichHandler()
        else:
            log_handler = logging.StreamHandler()

            # create formatter
            formatter = logging.Formatter('%(levelname)s:%(threadName)s:%(name)s:%(message)s')


            # add formatter to console handler
            log_handler.setFormatter(formatter)
        log_handler.setLevel(logging.DEBUG)

        # add console handler to logger
        logger.addHandler(log_handler)

    logging.getLogger("vaex").setLevel(logging.ERROR)  # default to higest level
    _set_log_level(vaex.settings.main.logging.error, logging.ERROR)
    _set_log_level(vaex.settings.main.logging.warning, logging.WARNING)
    _set_log_level(vaex.settings.main.logging.info, logging.INFO)
    _set_log_level(vaex.settings.main.logging.debug, logging.DEBUG)
    # VAEX_DEBUG behaves similar to VAEX_LOGGING_DEBUG, but has more effect
    DEBUG_MODE = os.environ.get('VAEX_DEBUG', '')
    if DEBUG_MODE:
        _set_log_level(DEBUG_MODE, logging.DEBUG)
    
setup()
